/*
GLDM - Graphical Liberty Display Manager

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software Foundation,
Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

*/

#include <SDL/SDL_image.h>
#include <SDL/SDL_rotozoom.h>
#include <algorithm>
#include <err.h>
#include "lunaticinterface.h"
#include "../authprovider.h"
#include "lunatic/userbox.h"
#include "lunatic/defines.h"
#include <sys/time.h>
#include "../utils.h"


//// xp-like interface code ////

#define SET_IMAGE_NULL(x) x = NULL;
#define UNLOAD_IMAGE(x) SDL_FreeSurface(x); x = NULL;

LunaticInterface::LunaticInterface()
{
    int ptsize=11;
    char fontname[]="DejaVuSans.ttf";

    TTF_Init();

    mFont = TTF_OpenFont(fontname,ptsize);
    if(mFont == NULL)
    {
        fprintf(stderr, "Couldn't load %d pt font from %s: %s\n", ptsize, fontname, SDL_GetError());
    }

    int renderstyle=0;
    //renderstyle |= TTF_STYLE_BOLD;
    TTF_SetFontStyle(mFont, renderstyle);






    SDL_EnableUNICODE(1);

    mSucceedingProvider = NULL;

    SET_IMAGE_NULL(mImgBG);
    SET_IMAGE_NULL(mImgBottom);
    SET_IMAGE_NULL(mImgDistLogo);
    SET_IMAGE_NULL(mImgGlow);
    SET_IMAGE_NULL(mImgLoginBox1);
    SET_IMAGE_NULL(mImgLoginBox1a);
    SET_IMAGE_NULL(mImgLoginBox2);
    SET_IMAGE_NULL(mImgQuitBtnNormal);
    SET_IMAGE_NULL(mImgSeparator);
    SET_IMAGE_NULL(mImgTop);

    mWelcomeFont =
    mInstructionFont =
    mMOTDFont = NULL;

    mMOTD = "After you log on, you can add or change accounts.\n"
            "Just go to Control Panel and click User Accounts.";


    mTurnOffBtnState = LunaticButton::NORMAL;
}
LunaticInterface::~LunaticInterface()
{
    unloadResources();
}

void LunaticInterface::unloadResources()
{

    UNLOAD_IMAGE(mImgBG);
    UNLOAD_IMAGE(mImgBottom);
    UNLOAD_IMAGE(mImgDistLogo);
    UNLOAD_IMAGE(mImgGlow);
    UNLOAD_IMAGE(mImgLoginBox1);
    UNLOAD_IMAGE(mImgLoginBox1a);
    UNLOAD_IMAGE(mImgLoginBox2);
    UNLOAD_IMAGE(mImgQuitBtnNormal);
    UNLOAD_IMAGE(mImgSeparator);
    UNLOAD_IMAGE(mImgTop);

    TTF_CloseFont(mWelcomeFont);
    TTF_CloseFont(mInstructionFont);
    TTF_CloseFont(mMOTDFont);
}

bool LunaticInterface::loadResources(const std::string& path)
{

    int renderstyle;
    mWelcomeFont = TTF_OpenFont((path+"Arial_Black.ttf").c_str(),WELCOME_SIZE);
    if(mWelcomeFont == NULL)
    {
        fprintf(stderr, "Couldn't load %d pt font from Arial_Black.tt: %s\n", WELCOME_SIZE, SDL_GetError());
        return false;
    }

    renderstyle=0;
    renderstyle |= TTF_STYLE_ITALIC;
    TTF_SetFontStyle(mWelcomeFont, renderstyle);



    mInstructionFont = TTF_OpenFont((path+"Arial.ttf").c_str(),INSTRUCTION_SIZE);
    if(mInstructionFont == NULL)
    {
        fprintf(stderr, "Couldn't load %d pt font from Arial.ttf: %s\n", INSTRUCTION_SIZE, SDL_GetError());
        return false;
    }
    renderstyle=0;
    //renderstyle |= TTF_STYLE_BOLD;
    TTF_SetFontStyle(mInstructionFont, renderstyle);


    mMOTDFont = TTF_OpenFont((path+"tahoma.ttf").c_str(),MOTD_SIZE);
    if(mMOTDFont == NULL)
    {
        fprintf(stderr, "Couldn't load %d pt font from tahoma.ttf: %s\n", MOTD_SIZE, SDL_GetError());
        return false;
    }
    renderstyle=0;
    TTF_SetFontStyle(mMOTDFont, renderstyle);



    mImgGlow = IMG_Load((path+"glow.png").c_str());                     if (!mImgGlow) return false;
    mImgDistLogo = IMG_Load((path+"distlogo.png").c_str());             if (!mImgDistLogo) return false;
    mImgLoginBox1 = IMG_Load((path+"loginbox1.png").c_str());           if (!mImgLoginBox1) return false;
    mImgLoginBox2 = IMG_Load((path+"loginbox2.png").c_str());           if (!mImgLoginBox2) return false;
    mImgLoginBox1a = IMG_Load((path+"loginbox1a.png").c_str());         if (!mImgLoginBox1a) return false;


    mImgQuitBtnNormal = IMG_Load((path+"quitbtn.png").c_str());         if (!mImgQuitBtnNormal) return false; // FIXME into lunaticbutton


    mBtnTurnOff = LunaticButton("turnoffbtn","png",path, 25, mScreen->h - 74);




    {
        SDL_Surface *s, *s2;

        s = IMG_Load((path+"bg.png").c_str()); if(!s) return false;

        s2 = zoomSurface(s, mScreen->w/(float)s->w, mScreen->h/(float)s->h, 0);
        mImgBG = SDL_DisplayFormat(s2);

        SDL_FreeSurface(s2);
        SDL_FreeSurface(s);
    }

    {
        SDL_Surface* s;

        s = IMG_Load((path+"top.png").c_str()); if(!s) return false;
        mImgTop = zoomSurface(s, mScreen->w/(float)s->w, 1, 0);
        SDL_FreeSurface(s);
    }
    {
        SDL_Surface* s;

        s = IMG_Load((path + "bottom.png").c_str()); if(!s) return false;
        mImgBottom = zoomSurface(s, mScreen->w/(float)s->w, 1, 0);
        SDL_FreeSurface(s);
    }

    {
        SDL_Surface* s;

        s = IMG_Load((path+"separator.png").c_str()); if(!s) return false;
        mImgSeparator = zoomSurface(s, 1, (mScreen->h - mImgBottom->h - mImgTop->h)/(float)s->h, 0);
        SDL_FreeSurface(s);
    }


    mRsrcPath=path;
    return true;
}
void LunaticInterface::showSelf()
{
    mEnteringLogin = true; mEnteringPassword = false;
    SDLInterface::showSelf();

    // check if video was inited
    if (!SDL_WasInit(SDL_INIT_VIDEO))
        errx(3, "Video was not init for %s\n", __PRETTY_FUNCTION__);


    // choose lunatic's theme
    std::string themename = "luna";
    FILE*f=fopen("lunatic.cfg","r");
    if(f)
    {
        char cthemename[255]={0};
        fscanf(f, "%s", cthemename);
        fclose(f);
        themename=cthemename;
    }

    // load lunatic's theme
    unloadResources();
    if (!loadResources(std::string("rsrc/") + themename + "/"))
    {
        unloadResources();
        if (!loadResources("rsrc/lunatic/"))
        {
            errx(4, "Resources not found for %s\n", __PRETTY_FUNCTION__);
        }
    }

    // let's get up user list!
    std::vector<std::string> usernames =AuthProvider::getLoginListUsers();
    sort(usernames.begin(), usernames.end());
    if(usernames.size()==0)
    {
        // same as in real lunatic's OS -- not even takin' root's username  :-)
        mUserList["root"] = LunaticUserbox("root", "Administrator", mRsrcPath, mScreen->w/2+10, mScreen->h/2 - mImgLoginBox1->h/2);
    }

    // let's set up user list!
    int totalSpace = mScreen->h - (mImgTop->h + mImgBottom->h);
    int requiredSpace = mImgLoginBox1->h * usernames.size();

    int current_y=mScreen->h/2 - requiredSpace/2;
    if(requiredSpace > totalSpace)
    {
        current_y = mImgTop->h;
    }

    for(std::vector<std::string>::iterator it = usernames.begin(); it != usernames.end(); it++)
    {
        mUserList[*it] = LunaticUserbox(*it, AuthProvider::getDisplayName(*it), mRsrcPath, mScreen->w/2+10, current_y);

        current_y += mImgLoginBox1->h;
    }

    mHoveredUser="";


}
void LunaticInterface::paintSelf()
{
    const SDL_VideoInfo* vi = SDL_GetVideoInfo();

    if (!SDL_WasInit(SDL_INIT_VIDEO))
        errx(3, "Video was not init for %s\n", __PRETTY_FUNCTION__);

	//SDL_Rect r={0,0,mScreen->w, mScreen->h};
	//SDL_FillRect(mScreen, &r, SDL_MapRGBA(vi->vfmt, 0x5a, 0x7e, 0xdb,255));

    blitSurface(mImgBG, 0,0);
    blitSurface(mImgGlow, 0,mImgTop->h);
    blitSurface(mImgTop, 0,0);
    blitSurface(mImgBottom, 0,mScreen->h - mImgBottom->h);

    bool mWelcome = false;
    if (mWelcome)
    {
        for (int i = 0; i < 2; i++){
            int size;
            mActiveFont=mWelcomeFont;
            mFontForeCol.r = mFontForeCol.g = mFontForeCol.b = 255 * i;
            mFontBackCol.r = 0x5a; mFontBackCol.g = 0x7e; mFontBackCol.b = 0xdb;
            mFontShaded=BLENDED;
            size=fontSize("welcome");
            //fontDrawText("welcome", mScreen->w/2 - size - 10 - i*4, mScreen->h/2 - 18 - i*4);
            fontDrawText("welcome", mScreen->w/2 - size/2 - i*4, mScreen->h/2 - WELCOME_SIZE/2 - i*4);
        }
    }
    else
    {
        blitSurface(mImgSeparator, mScreen->w/2, mImgTop->h);
        blitSurface(mImgDistLogo, mScreen->w/2 - mImgDistLogo->w - 10, mScreen->h/2 - mImgDistLogo->h);


        int size;
        std::string info = "To begin, click your user name";
        mActiveFont=mInstructionFont;
        mFontForeCol.r = mFontForeCol.g = mFontForeCol.b = 255;
        mFontBackCol.r = 0x5a; mFontBackCol.g = 0x7e; mFontBackCol.b = 0xdb;
        mFontShaded=BLENDED;
        size=fontSize(info.c_str());
        fontDrawText(info.c_str(), mScreen->w/2 - size - 28, mScreen->h/2 + INSTRUCTION_SIZE);


        SDL_Rect cliprect = {0, mImgTop->h, mScreen->w, mScreen->h - mImgTop->h - mImgBottom->h};
        SDL_SetClipRect(mScreen, &cliprect);

        for(std::map<std::string, LunaticUserbox>::iterator it = mUserList.begin(); it != mUserList.end(); it++)
        {
            std::string asterisks="";
            struct timeval tv;

            for(unsigned int i=0;i<getPassword().size();i++)
                asterisks+="*";
            gettimeofday(&tv, NULL);

            asterisks+=((tv.tv_sec*2 + (tv.tv_usec/500000)) % 2 ? "|" : "");

            it->second.paintSelf(this,
                                 (mEnteringPassword && mLogin==it->first ? mImgLoginBox2 : (mHoveredUser==it->first ? mImgLoginBox1a : mImgLoginBox1)), // bg
                                 ((mHoveredUser=="" && !mEnteringPassword) || mLogin==it->first || mHoveredUser==it->first) ? 0.0f:0.6f, // dimmed
                                 (mEnteringPassword && mLogin==it->first ? "Type your password" : ""), // instruction
                                 (mEnteringPassword && mLogin==it->first ? asterisks : ""));

        }
        //paintUserBox("ivucica", mScreen->w/2 + 10, mScreen->h/2 - mImgLoginBox1->h/2);

        SDL_SetClipRect(mScreen, NULL);


        // paint turn off computer btn
        mActiveFont=mInstructionFont;
        if(mTurnOffBtnState != LunaticButton::NORMAL)
            TTF_SetFontStyle(mInstructionFont, TTF_STYLE_UNDERLINE);
        mFontForeCol.r=mFontForeCol.g=mFontForeCol.b=255;
        fontDrawText("Turn off computer", 55, mScreen->h - 74);
        TTF_SetFontStyle(mInstructionFont, 0);

        mBtnTurnOff.paintSelf(this, mTurnOffBtnState);


        // paint motd
        mActiveFont=mMOTDFont;
        mFontForeCol.r=mFontForeCol.g=mFontForeCol.b=255;
        fontDrawText(mMOTD.c_str(), mScreen->w-331, mScreen->h-67);

    }


    SDL_Flip(mScreen);

}


void LunaticInterface::updateSelf(double k)
{
    SDL_Event event;
    while(SDL_PollEvent(&event)){
        switch (event.type){
            case SDL_KEYDOWN:
            if (mEnteringLogin)
            {
                switch (event.key.keysym.unicode) {
                    case 8:
                        mLogin=mLogin.substr(0,mLogin.size()-1);
                        break;
                    case 13:
                    {
                        mEnteringLogin = false;
                        mEnteringPassword = true;
                        break;
                    }
                    case 27:
                        mPerformQuit=true;
                        break;
                    default:
                    {
                        if (isalnum(event.key.keysym.unicode))
                            mLogin += event.key.keysym.unicode;
                        break;
                    }
                }
            }
            else if (mEnteringPassword)
            {
                switch (event.key.keysym.unicode) {
                    case 8:
                        mPassword=mPassword.substr(0,mPassword.size()-1);
                        break;
                    case 13:
                    {
                        if((mSucceedingProvider=AuthProvider::checkAuth(getLogin(), getPassword())))
                        {
                            mEnteringLogin = false;
                            mEnteringPassword = false;
                        } else {
                            warnx("incorrect login\n");
                            mLogin = mPassword = "";
                            mEnteringLogin = true; mEnteringPassword = false;
                        }
                        break;
                    }
                    case 27:
                        mPerformQuit=true;
                        break;
                    default:
                    {
                        if (isalnum(event.key.keysym.unicode))
                            mPassword += event.key.keysym.unicode;
                        break;
                    }
                }
            }
            break;
            case SDL_MOUSEMOTION:
            {

                int x=event.motion.x, y=event.motion.y;

                int min_x=mScreen->w/2 + 10;
                int max_x=mScreen->w/2 + 10 + mImgLoginBox2->w;
                int min_y=mScreen->h/2 - (mImgLoginBox1->h * mUserList.size())/2;;
                int max_y=mScreen->h/2 + (mImgLoginBox1->h * mUserList.size())/2;



                int totalSpace = mScreen->h - (mImgTop->h + mImgBottom->h);
                int requiredSpace = mImgLoginBox1->h * mUserList.size();

                if(requiredSpace > totalSpace)
                {
                    min_y=mImgTop->h;
                    max_y=mScreen->h-mImgBottom->h;
                }

                mHoveredUser="";
                if(x > min_x && x < max_x && y > min_y && y < max_y)
                {
                    for(std::map<std::string,LunaticUserbox>::iterator it = mUserList.begin(); it != mUserList.end(); it++)
                    {
                        if(it->second.pointInside(this,x,y))
                        {
                            mHoveredUser=it->first;
                        }
                    }
                }


                // if it's in area enclosing both the button and the label, then it's hilited
                mActiveFont=mInstructionFont;
                if(x > 25               && x <= 55 + fontSize("Turn off computer") &&
                   y > mScreen->h - 74  && y <= mScreen->h - 74 + 26 )
                {
                    mTurnOffBtnState=LunaticButton::HOVER;
                }
                else
                {
                    mTurnOffBtnState=LunaticButton::NORMAL;
                }



            }
            break;
            case SDL_MOUSEBUTTONDOWN:
            {
                if(mHoveredUser != "")
                {
                    mLogin = mHoveredUser;
                    mEnteringLogin=false;
                    mEnteringPassword=true;
                }

                break;
            }
            default:
            handleEvent(k, event);
        }
    }


    for(std::map<std::string, LunaticUserbox>::iterator it = mUserList.begin(); it != mUserList.end(); it++)
    {
        it->second.updateSelf(k);
    }

}
bool LunaticInterface::isFinished()
{
    return !mEnteringLogin && !mEnteringPassword;
}
bool LunaticInterface::isCanceled()
{
    return SDLInterface::isCanceled();
}


const std::string& LunaticInterface::getLogin() const
{
    return mLogin;
}
const std::string& LunaticInterface::getPassword() const
{
    return mPassword;
}


//// sdl font code ////

int LunaticInterface::fontDrawChar(char character, int x, int y)
{

    // solid or shaded; shaded uses backcol, solid doesnt
    // blended also doesnt have bg color, but is also antialiasing
    SDL_Surface* glyph;
    switch(mFontShaded)
    {
    case SOLID: glyph = TTF_RenderGlyph_Solid( mActiveFont, character, mFontForeCol); break;
    case SHADED: glyph = TTF_RenderGlyph_Shaded( mActiveFont, character, mFontForeCol, mFontBackCol); break;
    case BLENDED: glyph = TTF_RenderGlyph_Blended( mActiveFont, character, mFontForeCol); break;
    default: return 0;
    }

    int minx, maxx, miny, maxy, advance;

    TTF_GlyphMetrics(mActiveFont, character, &minx, &maxx, &miny, &maxy, &advance);

    SDL_Rect src = {0, 0, glyph->w, glyph->h};
    SDL_Rect dest = {x+minx,y+TTF_FontAscent(mActiveFont)-maxy, glyph->w, glyph->h};

    SDL_BlitSurface(glyph, &src, mScreen, &dest);

    SDL_FreeSurface(glyph);

    return advance;
}
void LunaticInterface::fontDrawText(std::string txt_s, float x, float y)
{
    std::vector<std::string> lines = StrExplode(txt_s, "\n");

    for(std::vector<std::string>::iterator it = lines.begin(); it != lines.end(); it++)
    {
        int yoffset=(it-lines.begin())*TTF_FontLineSkip(mActiveFont);

        SDL_Surface *text=0;
        switch(mFontShaded)
        {
        case SOLID:  text=TTF_RenderUTF8_Solid(mActiveFont, it->c_str(), mFontForeCol); break;
        case SHADED: text=TTF_RenderUTF8_Shaded( mActiveFont, it->c_str(), mFontForeCol, mFontBackCol); break;
        case BLENDED: text=TTF_RenderUTF8_Blended( mActiveFont, it->c_str(), mFontForeCol); break;
        default: return ;
        }
        if(!text)return;

        SDL_Rect src = {0, 0, text->w, text->h};
        SDL_Rect dest = {x,y+yoffset, text->w, text->h};

        SDL_BlitSurface(text, &src, mScreen, &dest);

        SDL_FreeSurface(text);

    }

}
float LunaticInterface::fontSize(std::string txt)
{
    int sum=0;
    int advance;
    for(unsigned int i=0;i<txt.size();i++)
    {

        TTF_GlyphMetrics(mActiveFont, txt[i], NULL,NULL,NULL,NULL, &advance);
        sum+=advance;
    }
    return sum;
}
