/*
GLDM - Graphical Liberty Display Manager

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software Foundation,
Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

*/

#include "simplesdlinterface.h"
#include "../authprovider.h"
#include <err.h>

//// simple sdl interface code ////


SimpleSDLInterface::SimpleSDLInterface()
{
    int ptsize=11;
    char fontname[]="DejaVuSans.ttf";

    TTF_Init();

    mFont = TTF_OpenFont(fontname,ptsize);
    if(mFont == NULL)
    {
        fprintf(stderr, "Couldn't load %d pt font from %s: %s\n", ptsize, fontname, SDL_GetError());
    }

    int renderstyle=0;
    renderstyle |= TTF_STYLE_BOLD;
    TTF_SetFontStyle(mFont, renderstyle);

    SDL_EnableUNICODE(1);

    mSucceedingProvider = NULL;

}
SimpleSDLInterface::~SimpleSDLInterface()
{
}

void SimpleSDLInterface::showSelf()
{
    mEnteringLogin = true; mEnteringPassword = false;
    SDLInterface::showSelf();
}
void SimpleSDLInterface::paintSelf()
{
    const SDL_VideoInfo* vi = SDL_GetVideoInfo();

    if (!SDL_WasInit(SDL_INIT_VIDEO))
        errx(3, "Video was not init for %s\n", __PRETTY_FUNCTION__);
	SDL_Rect r={0,0,mScreen->w, mScreen->h};
	SDL_FillRect(mScreen, &r, SDL_MapRGBA(vi->vfmt, 0,0,0,255));

    if(mEnteringLogin)
    {
        fontDrawText("login: ", 50, 50);
        fontDrawText(getLogin(), 50+fontSize("login: ") ,50);
    } else if (mEnteringPassword)
    {
        std::string asterisks="";
        for(unsigned int i=0;i<getPassword().size();i++)
            asterisks+="*";
        fontDrawText("password: ", 50, 50);
        fontDrawText(asterisks, 50+fontSize("password: ") ,50);
    }
    SDL_Flip(mScreen);

}

void SimpleSDLInterface::updateSelf(double k)
{
    SDL_Event event;
    while(SDL_PollEvent(&event)){
        switch (event.type){
            case SDL_KEYDOWN:
            if (mEnteringLogin)
            {
                switch (event.key.keysym.unicode) {
                    case 8:
                        mLogin=mLogin.substr(0,mLogin.size()-1);
                        break;
                    case 13:
                    {
                        mEnteringLogin = false;
                        mEnteringPassword = true;
                        break;
                    }
                    case 27:
                        mPerformQuit=true;
                        break;
                    default:
                    {
                        if (isalnum(event.key.keysym.unicode))
                            mLogin += event.key.keysym.unicode;
                        break;
                    }
                }
            }
            else if (mEnteringPassword)
            {
                switch (event.key.keysym.unicode) {
                    case 8:
                        mPassword=mPassword.substr(0,mPassword.size()-1);
                        break;
                    case 13:
                    {
                        if((mSucceedingProvider=AuthProvider::checkAuth(getLogin(), getPassword())))
                        {
                            mEnteringLogin = false;
                            mEnteringPassword = false;
                        } else {
                            mLogin = mPassword = "";
                            mEnteringLogin = true; mEnteringPassword = false;
                        }
                        break;
                    }
                    case 27:
                        mPerformQuit=true;
                        break;
                    default:
                    {
                        if (isalnum(event.key.keysym.unicode))
                            mPassword += event.key.keysym.unicode;
                        break;
                    }
                }
            }
            break;
            default:
            handleEvent(k, event);
        }
    }


}
bool SimpleSDLInterface::isFinished()
{
    return !mEnteringLogin && !mEnteringPassword;
}
bool SimpleSDLInterface::isCanceled()
{
    return SDLInterface::isCanceled();
}


const std::string& SimpleSDLInterface::getLogin() const
{
    return mLogin;
}
const std::string& SimpleSDLInterface::getPassword() const
{
    return mPassword;
}


//// sdl font code ////

int SimpleSDLInterface::fontDrawChar(char character, int x, int y)
{
    SDL_Color forecol, backcol;

    forecol.r = 255;
    forecol.g = 255;
    forecol.b = 255;

    backcol.r = 0;
    backcol.g = 0;
    backcol.b = 0;


    SDL_Surface* glyph = TTF_RenderGlyph_Solid( mFont, character, forecol); // solid or shaded; shaded uses backcol, solid doesnt
    int minx, maxx, miny, maxy, advance;

    TTF_GlyphMetrics(mFont, character, &minx, &maxx, &miny, &maxy, &advance);

    SDL_Rect src = {0, 0, glyph->w, glyph->h};
    SDL_Rect dest = {x+minx,y+TTF_FontAscent(mFont)-maxy, glyph->w, glyph->h};

    SDL_BlitSurface(glyph, &src, mScreen, &dest);

    SDL_FreeSurface(glyph);

    return advance;
}
void SimpleSDLInterface::fontDrawText(std::string txt_s, float x, float y)
{
    const char* txt = txt_s.c_str();
    float cx=x,cy=y;

    float sizesofar = 0.;
    float linessofar = 0.;
    for (unsigned char *t = (unsigned char*)txt; *t; ++t)
    {
        switch (*t)
        {
            default:
            {
                int advance = fontDrawChar(*t, cx, cy);
                cx += advance;
                sizesofar += advance;
                break;
            }

            case '\n':
            case '\r':
            {
                cx -= sizesofar;
                cy += 10;
                linessofar += 1.;
                sizesofar = 0;
                if ((*t == '\n' && *(t+1)=='\r') || (*t == '\r' && *(t+1)=='\n'))
                    t++;
            }
            break;

        }
    }
}
float SimpleSDLInterface::fontSize(std::string txt)
{
    int sum=0;
    int advance;
    for(unsigned int i=0;i<txt.size();i++)
    {

        TTF_GlyphMetrics(mFont, txt[i], NULL,NULL,NULL,NULL, &advance);
        sum+=advance;
    }
    return sum;
}

